{{
********************************
* jvmNative - native functions *
********************************
}}


OBJ
  comm:  "jvmComm"  'communication port object
  jd:    "jvmData"  'jvm memory access object

  
PUB doNative(nm_function)
{{
   Execute a native function. Arguments for native functions are either in VP registers
   or on the stack. If a native function returns a value it must be placed on top of the stack.
}}
  case nm_function
     0: jd.writeRegister(0,0)
     1: jd.readRegister(0)
     2: nm_core_CPU_writeObject(0,0)
     3: nm_core_CPU_delay(0)
     4: comm.message(0,0)
     5: nm_page_
     6: nm_core_CPU_count(0,0,0)
     7: nm_core_CPU_rcTime(0,0,0)
     8: nm_core_CPU_pulseIn(0,0,0)
     9: nm_core_CPU_pulseOut(0,0)
    10: nm_core_CPU_readPin(0)
    11: nm_core_CPU_writePin(0,0)
    12: nm_core_CPU_shiftIn(0,0,0,0)
    13: nm_core_CPU_shiftOut(0,0,0,0,0)
    14: jd.isCarry
    15: nm_page_
    16: nm_core_CPU_uninstall(0)
    17: nm_core_CPU_install(0)
    18: nm_core_CPU_nap(0)
    19: nm_core_CPU_setOutput(0)
    20: nm_page_
    21: nm_core_CPU_setInput(0)
    22: nm_page_
    23: nm_core_CPU_readSRAM(0)
    24: nm_page_
    25: nm_core_CPU_writeSRAM(0,0)
    26: nm_page_
    27: nm_core_EEPROM_eeRead(0)
    28: nm_page_
    29: nm_core_EEPROM_eeWrite(0,0)
    30: nm_page_
    31: nm_core_Uart_txInit(0)
    32: nm_core_Uart_rxInit(0)
    33: nm_core_Uart_rxRead(0)
    34: nm_core_Callback_returnTwo
    35: nm_core_Timer_init(0)
    36: nm_core_Timer_latch(0)
    37: nm_page_
    38: nm_core_PWM_install(0)
    39: nm_page_
    40: nm_core_PWM_updateInternal(0,0,0,0,0)
    41: nm_page_
    42: nm_core_DAC_install(0)
    43: nm_page_
    44: nm_core_ADC_install(0)
    45: nm_page_
    46: return comm.getByte
    47: nm_page_


PRI nm_core_CPU_writeObject(address,value)
  'Needs to write where ???


PRI nm_core_CPU_delay(period) | t
  '/**
  ' * Delays the program execution for a period of time. The delay method does not
  ' * return until the delay period has expired. During a delay any installed Virtual
  ' * Peripherals will continue to run.
  ' * <p>
  ' *
  ' * @param period the length of time to delay for, measured in
  ' * 95.48 microsecond units.
  ' */
  t := cnt
  repeat until (cnt - t) / (clkfreq * 95.48 / 1000000) > period

  
PRI nm_core_CPU_count(timeout,portPin,edge): value | pin,t
  '/**
  ' * Count the number of rising or falling edge transitions that occur during
  ' * a set period of time.
  ' * If the pin is an output then it will be changed to be an input and will remain as
  ' * an input when the counting is complete.
  ' * <p>
  ' * To ensure reliable detection the input signal must maintain a value for at least
  ' * 8.68us. This means that a square wave must have a period of at least 17.36us and
  ' * a frequency less than 57.6kHz.
  ' *
  ' * @param timeout the amount of time to count transitions over. Measured in
  ' *                8.68us units. The value of 32,767 allows a maximum measurement
  ' *                time of 284.4ms.
  ' * @param portPin     the I/O pin to measure transitions on.
  ' * @param edge    the type of edge to count. Pass true to count rising edges,
  ' *                or false to count falling edges.
  ' * @return the number of rising edges or falling edges that occured on the pin
  ' *         during the timeout period.
  ' */
  {{ adapted from object BS2_functions PUB COUNT (Pin, Duration) : Value }}
  {{ only does rising-edge pulses now  }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  dira[pin]~                                          ' Set as input
  ctra := 0                                           ' Clear any value in ctrb
                                                           ' set up counter, pos edge trigger
  ctra := (%01010 << 26 ) | (%001 << 23) | (0 << 9) | (pin)  
  frqa := 1                                           ' 1 count/trigger
  phsa := 0                                           ' Clear phase - holds accumulated count
  t := cnt
  repeat until (cnt - t) / (clkfreq * 8.68 / 1000000) > timeout ' Allow to count for duration in 8.68us units 
  value := phsa                                       ' Return total count
  ctra := 0                                           ' clear counter mode


PRI nm_core_CPU_rcTime(timeout,portPin,pinState): value | pin, ClkStart,clkStop
  '/**
  ' * Measure the length of time it takes for a pin to match a desired value.
  ' * If the pin is an output then it will be changed to be an input and will
  ' * remain as an input when rcTime is complete. rcTime can be used to measure
  ' * the charge/discharge time of a RC (resistor-capacitor) circuit in order to
  ' * read an analogue value.
  ' *
  ' * @param timeout the maximum measurable time.
  ' * @param portPin the I/O pin to measure.
  ' * @param pinState the pin state to wait for.
  ' * @return the amount of time taken for the pin to change to the desired value,
  ' * measured in 8.68us units. A value of -1 indicates that the timeout expired
  ' * before the desired state was reached.
  ' */
  {{ adapted from object BS2_functions PUB RCTIME (Pin,State):Duration | ClkStart, ClkStop }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  if pinState
    pinState := 1
  else
    pinState := 0
  dira[pin]~
  ClkStart := cnt                                         ' Save counter for start time
  waitpne(pinState << pin, |< pin, 0)                     ' Wait for opposite state to end
  clkStop := cnt                                          ' Save stop time
  value := (clkStop - ClkStart)/(clkfreq * 8.68 /1000000) ' calculate in 8.68us resolution


PRI nm_core_CPU_pulseIn(timeout,portPin,pinState): value | pin
  '/**
  ' * Measure the length of a single pulse. Waits for the value of the selected
  ' * pin to match the value selected by the state parameter and then times how
  ' * long it takes for the pin to return to its original state.
  ' * If the pin is an output then it will be changed to be an input and will
  ' * remain as an input when complete.
  ' * <p>
  ' * A pulse of less than 8.68us in duration may be missed.
  ' *
  ' * @param timeout  the maximum amount of time to spend in the pulseIn method.
  ' *                 The timeout covers the complete time of the process both
  ' *                 the wait for the start of the pulse and the time of the pulse.
  ' * @param portPin      the I/O pin to measure the pulse on.
  ' * @param pinState    whether to measure a high pulse (true) or a low pulse (false).
  ' * @return the length of the pulse in 8.68us units. A value of 0 indicates
  ' *         that no start edge was detected. A value of -1 indicates that no
  ' *         stop edge was detected.
  ' */
  {{ adapted from object BS2_functions PUB PULSIN_Clk(Pin, State) : Duration }}
  {{ need to implement timeout }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  dira[pin]~
  ctra := 0
  if pinState
    pinState := 1
  else
    pinState := 0
  if pinState == 1
    ctra := (%11010 << 26 ) | (%001 << 23) | (0 << 9) | (pin) ' set up counter, A level count
  else
    ctra := (%10101 << 26 ) | (%001 << 23) | (0 << 9) | (pin) ' set up counter, !A level count
  frqa := 1
  waitpne(pinState << pin, |< Pin, 0)                         ' Wait for opposite state ready
  phsa:=0                                                  ' Clear count
  waitpeq(pinState << pin, |< Pin, 0)                         ' wait for pulse
  waitpne(pinState << pin, |< Pin, 0)                         ' Wait for pulse to end
  value := phsa * 1000000 /clkfreq / 8.68                  ' Return duration as 8.68us units
  ctra :=0                                                 ' stop counter


PRI nm_core_CPU_pulseOut(length,portPin) | ClkCycles,pin
  '/**
  ' * Generate a pulse of a specific duration. If the pin is an input then it will
  ' * be changed to be an output and remain as an output when complete. To
  ' * ensure that a positive going pulse or negative going pulse is generated it
  ' * is a good idea to write a value to the pin first to put it into the correct
  ' * state. pulseOut inverts the value of the selected pin, waits for the
  ' * specified length of time, and then returns the pin to its original value.
  ' * <p>
  ' * The time is measured in 8.68us units with a maximum value of 32767 giving a
  ' * maximum pulse time of 284.4ms. A time of 0 will not cause a pulse.
  ' * pulseOut can be used to generate the control pulses for a servo motor.
  ' *
  ' * @param length the length of the pulse to generate, measured in 8.68us units.
  ' * @param portPin    the I/O to generate the pulse on.
  ' */
  {{ adapted from object BS2_functions PUB PULSOUT(Pin,Duration)  | clkcycles }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  ClkCycles := (length * clkfreq / 1000000 * 8.68 - 1250) #> 400  ' 400 = Minimum waitcnt value to prevent lock-up
                       ' duration * clk cycles for 8.68us
                                                           ' - inst. time, min cntMin
  dira[pin]~~                                              ' Set to output                                         
  !outa[pin]                                               ' set to opposite state
  waitcnt(clkcycles + cnt)                                 ' wait until clk gets there 
  !outa[pin]                                               ' return to orig. state


PRI nm_core_CPU_readPin(portPin): YesNo
  '/**
  ' * Reads the logic value on a pin. If the pin is an output then it will be
  ' * changed to be an input and will remain as an input when complete.
  ' *
  ' * @param portPin the I/O pin to read.
  ' * @return the state of the pin where true indicates high and false indicates low.
  ' */
  if portPin < 256
    dira[7..0] &= (255-portPin)
    return ina[>|portPin]
  else
    portPin -=256
    dira[15..8] &= (255-portPin)
    return ina[8 + >|portPin]
    

PRI nm_core_CPU_writePin(portPin,value)
  '/**
  ' * Set a pin to a logic state. Makes the selected pin high or low as selected
  ' * by the state parameter. If the pin is an input then it will be changed to
  ' * be an output and remain as an output when complete.
  ' *
  ' * @param portPin the I/O pin to write.
  ' * @param value the state to set the pin to (true=high, false=low).
  ' */
  if portPin < 256
    dira[7..0] |= portPin
    outa[>|portPin] := value
  else
    portPin -=256
    dira[15..8] |= portPin
    outa[8 + >|portPin] := value


PRI nm_core_CPU_shiftIn(dataPortPin,clockPortPin,bitCount,mode): value | Dpin,Cpin,InBit
  '/**
  ' * Read data from a synchronous serial device.
  ' * The clock pin outputs a clock signal and the data is clocked in form the data pin.
  ' * If the data pin is an output then it will be changed to be an input and remain as
  ' * an input when complete. If the clock pin is an input then it will be changed to be
  ' * an output and remain as an output when complete.
  ' * <p> The mode parameter selects pre/post clock sampling. Note that some devices
  ' * require pre-clock sampling. When pre-clock sampling is used the data pin is
  ' * sampled to obtain the first bit prior to any change of the clock pin. Clearly
  ' * the device must have had a clock pulse to output the first bit and this is
  ' * normally supplied by the previous call to shiftIn or shiftOut. Most
  ' * synchronous serial devices require a write (shiftOut) prior to reading and so the
  ' * clock for the first bit will be supplied by the shiftOut call. If the device must be
  ' * read without first being written to then the first read (shiftIn) should request an
  ' * extra bit to provide the initial clock pulse.
  ' * <p>
  ' * Note that the bit count can be set to higher than 16 (up to 256) if desired. When
  ' * a number greater than 16 is used the last 16 bits will be returned.
  ' *
  ' * @param dataPortPin the I/O pin to read data on.
  ' * @param clockPortPin the I/O pin on which to provide the clock to the external device.
  ' * @param bitCount the number of bits to count in (from 1 to 16). bitCount must not be 0.
  ' * @param mode the shifting and clocking mode to employ. This determines whether
  ' * bits are shifted into the least-significant-bit (LSb) or most-significant-bit (MSb)
  ' * first. The clocking mode determines whether the data is sampled before or after
  ' * the first clock transition. See the description for the possible values.
  ' * @return the bits shifted in on the data pin. The result contains bitCount bits and is
  ' * justified according to the mode setting. If MSb first is selected then the result
  ' * will be the data read in on the port. If LSb first is selected then the data is
  ' * returned MSb justified and must be shifted right (16-bitCount) bits to
  ' * obtain the correct value.
  ' */
  {{ adapted from object BS2_functions PUB SHIFTIN (Dpin, Cpin, Mode, Bits) : Value | InBit }}
{{
   Shift data in, master clock, for mode use BS2#MSBPRE, #MSBPOST, #LSBPRE, #LSBPOST
   Clock rate is ~16Kbps.  Use at 80MHz only is recommended.
     X := BS2.SHIFTIN(5,6,BS2#MSBPOST,8)
}}
  if dataPortPin < 256
    Dpin := >|dataPortPin
  else
    dataPortPin &= $FF
    Dpin := 8 + >|dataPortPin
  if clockPortPin < 256
    Cpin := >|clockPortPin
  else
    clockPortPin &= $FF
    Cpin := 8 + >|clockPortPin
    dira[Dpin]~                                            ' Set data pin to input
    outa[Cpin]:=0                                          ' Set clock low 
    dira[Cpin]~~                                           ' Set clock pin to output 
                                                
  If Mode == jd#PRE_CLOCK_MSB                               ' Mode - MSB, before clock
       Value:=0
       REPEAT Bitcount                                         ' for number of bits
          InBit:= ina[Dpin]                                ' get bit value
          Value := (Value << 1) + InBit                    ' Add to  value shifted by position
          !outa[Cpin]                                      ' cycle clock
          !outa[Cpin]
          waitcnt(1000 + cnt)                              ' time delay

  elseif Mode == jd#POST_CLOCK_MSB                            ' Mode - MSB, after clock              
       Value:=0                                                          
       REPEAT Bitcount                                         ' for number of bits                    
          !outa[Cpin]                                      ' cycle clock                         
          !outa[Cpin]                                         
          InBit:= ina[Dpin]                                ' get bit value                          
          Value := (Value << 1) + InBit                    ' Add to  value shifted by position                                         
          waitcnt(1000 + cnt)                              ' time delay                            
                                                                 
  elseif Mode == jd#POST_CLOCK_LSB                            ' Mode - LSB, after clock                    
       Value:=0                                                                                         
       REPEAT Bitcount                                         ' for number of bits                         
          !outa[Cpin]                                      ' cycle clock                          
          !outa[Cpin]                                                                             
          InBit:= ina[Dpin]                                ' get bit value                        
          Value := (InBit << (bitcount-1)) + (Value >> 1)      ' Add to  value shifted by position    
          waitcnt(1000 + cnt)                              ' time delay                           

  elseif Mode == jd#PRE_CLOCK_LSB                             ' Mode - LSB, before clock             
       Value:=0                                                                                   
       REPEAT Bitcount                                         ' for number of bits                   
          InBit:= ina[Dpin]                                ' get bit value                        
          Value := (Value >> 1) + (InBit << (bitcount-1))      ' Add to  value shifted by position    
          !outa[Cpin]                                      ' cycle clock                          
          !outa[Cpin]                                                                             
          waitcnt(1000 + cnt)                              ' time delay

  elseif Mode == jd#OnClock  {{ is this required ?? }}                                          
       Value:=0
       REPEAT Bitcount                                         ' for number of bits
                                        
          !outa[Cpin]                                      ' cycle clock
          waitcnt(500 + cnt)                               ' get bit value
          InBit:= ina[Dpin]                               ' time delay
          Value := (Value << 1) + InBit                    ' Add to  value shifted by position
          !outa[Cpin]
          waitcnt(500 + cnt)                           
     



PRI nm_core_CPU_shiftOut(dataPortPin,clockPortPin,bitCount,mode,data) | Dpin,Cpin
  '/**
  ' * Write data to a synchronous serial device.
  ' * The clock pin outputs a clock signal and the data is clocked out the data pin. If
  ' * the data pin is an input then it will be changed to be an output and remain as an
  ' * output when complete. Likewise if the clock pin is an input then it will be
  ' * changed to be an output and remain as an output when complete.
  ' * <p>
  ' * Both the clock and data pins should be initialized before calling shiftOut. shiftOut
  ' * works by inverting the pins so you can generate either idle high or idle low signals.
  ' * <p>
  ' * If LSb first is selected then the data can be passed directly. If MSb first is
  ' * selected then if less than 16 bits are to be shift out, data should be shifted left
  ' * to MSb justify it.
  ' * @param dataPortPin the I/O pin to write data on.
  ' * @param clockPortPin the I/O pin on which to provide the clock to the external device.
  ' * @param bitCount the number of bits to count out (from 1 to 16). bitCount must not be 0.
  ' * @param mode the shifting mode to employ. This determines whether
  ' * bits are shifted into the least-significant-bit (LSb) or most-significant-bit (MSb)
  ' * first. Valid values are CPU.SHIFT_MSB and CPU.SHIFT_LSB.
  ' */
  {{ adapted from object BS2_functions PUB SHIFTOUT (Dpin, Cpin, Value, Mode, Bits)| bitNum  }}
{{
   Shift data out, master clock, for mode use ObjName#LSBFIRST, #MSBFIRST
   Clock rate is ~16Kbps.  Use at 80MHz only is recommended.
     BS2.SHIFTOUT(5,6,"B",BS2#LSBFIRST,8)
}}
  if dataPortPin < 256
    Dpin := >|dataPortPin
  else
    dataPortPin &= $FF
    Dpin := 8 + >|dataPortPin
  if clockPortPin < 256
    Cpin := >|clockPortPin
  else
  clockPortPin &= $FF
  Cpin := 8 + >|clockPortPin
  outa[Dpin]:=0                                          ' Data pin = 0
  dira[Dpin]~~                                           ' Set data as output
  outa[Cpin]:=0
  dira[Cpin]~~

  If Mode == jd#SHIFT_LSB                                    ' Send LSB first    
       REPEAT Bitcount
          outa[Dpin] := data                              ' Set output
          data := data >> 1                              ' Shift value right
          !outa[Cpin]                                      ' cycle clock
          !outa[Cpin]
          waitcnt(1000 + cnt)                              ' delay

  elseIf Mode == jd#SHIFT_MSB                                ' Send MSB first               
       REPEAT Bitcount                                                                
          outa[Dpin] := data >> (bitcount-1)                  ' Set output           
          data := data << 1                              ' Shift value right    
          !outa[Cpin]                                      ' cycle clock          
          !outa[Cpin]                                                             
          waitcnt(1000 + cnt)                              ' delay                
    outa[Dpin]~                                            ' Set data to low
     

  
PRI nm_core_CPU_uninstall(bank)
  '/**
  ' * This function is called by CPU.removeVP. 
  ' * Remove a virtual peripheral from its slot. The VP will stop executing
  ' * immediately and its slot will be available to install a different VP.
  ' *
  ' * @param bank the rambank allocated to the virtual peripheral that must stop.
  ' */


PRI nm_core_CPU_install(routine)
  'Appears not to be used

  
PRI nm_core_CPU_nap(period)
  'If the Prop cannot support nap, then just do a delay
  '/**
  ' * Put the Prop into sleep mode for a period of time. The accuracy
  ' * of the sleep timer is very poor. See the manual for more information.
  ' *
  ' * @param period specifies the period for which the Prop should sleep. The
  ' * following values are valid:
  ' *
  ' * 0 - 16ms
  ' * 1 - 32ms
  ' * 2 - 64ms
  ' * 3 - 128ms
  ' * 4 - 256ms
  ' * 5 - 512ms
  ' * 6 - 1024ms
  ' * 7 - 2048ms
  ' */


PRI nm_core_CPU_setOutput(portPin)
  '/**
  ' * Make a pin an output.
  ' * The specified pin will be converted to an output. More than one pin can
  ' * be specified using the | operator, as long as all pins are on the same
  ' * port.
  ' *
  ' * @param portPin the pin(s) to make into an output.
  ' */
  if portPin < 256
    dira[7..0] |= portPin
  else
    portPin &= $FF
    dira[15..8] |= portPin
    

PRI nm_core_CPU_setInput(portPin)
  '/**
  ' * Make a pin an input.
  ' * The specified pin will be converted to an input. More than one pin can
  ' * be specified using the | operator, as long as all pins are on the same
  ' * port.
  ' *
  ' * @param portPin the pin(s) to make into an input.
  ' */
  if portPin < 256
    dira[7..0] &= (255 ^ portPin)
  else
    portPin &= $FF
    dira[15..8] &= (255 ^ portPin)


PRI nm_core_CPU_readSRAM(address): value
  '/**
  ' * Reads a byte from SRAM at the specified address.
  ' *
  ' * @param address the address to read the byte from.
  ' * @return the byte read from the SRAM.
  ' */
  return byte[address]
  

PRI nm_core_CPU_writeSRAM(address,value)
  '/**
  ' * Writes a byte into the SRAM at the specified address.
  ' *
  ' * @param address the address to write the byte to.
  ' * @param value value to write.
  ' */
  byte[address] := value
  

PRI nm_core_EEPROM_eeRead(address): value | ack,temp
  '/**
  ' * Reads a byte from the EEPROM at the specified address.
  ' *
  ' * @param address the address to read the byte from.
  ' * @return the byte read from the EEPROM.
  ' */
  {{ adapted from object BS2_functions Pub Read(address, size) }}
  value~
  dira[comm#propSCL]~~
  outa[comm#propSCL]~~
  dira[comm#propSDA]~
  outa[comm#propSDA]~
  dira[comm#propSDA]~~
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, %10100000, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL, jd#PRE_CLOCK_LSB, 1)
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, address >> 8, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL, jd#PRE_CLOCK_LSB, 1)
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, address, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL, jd#PRE_CLOCK_LSB, 1)
  dira[comm#propSCL]~~
  outa[comm#propSCL]~~
  dira[comm#propSDA]~
  outa[comm#propSDA]~
  dira[comm#propSDA]~~
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, %10100001, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL, jd#PRE_CLOCK_LSB, 1)
  dira[comm#propSDA]~
  value := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL, jd#PRE_CLOCK_MSB, 8)
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL, jd#PRE_CLOCK_LSB, 1)
  outa[comm#propSDA]~
  dira[comm#propSDA]~~
  outa[comm#propSCL]~~
  dira[comm#propSDA]~
        

PRI nm_core_EEPROM_eeWrite(address,value) | ack 
  '/**
  ' * Writes a byte into the EEPROM at the specified address.
  ' *
  ' * @param address the address to write the byte to.
  ' * @param value value to write.
  ' */
  {{ adapted from object BS2_functions Pub Write(address, value, size) }}
  outa[comm#propSCL]~~
  dira[comm#propSCL]~~
  dira[comm#propSDA]~
  outa[comm#propSDA]~
  dira[comm#propSDA]~~
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, %10100000, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL,jd#OnClock,1)
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, address >> 8, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL,jd#OnClock, 1)
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, address, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL,jd#OnClock,1)
  nm_core_CPU_shiftout(comm#propSDA, comm#propSCL, value, jd#SHIFT_MSB, 8)
  dira[comm#propSDA]~
  ack := nm_core_CPU_shiftin(comm#propSDA, comm#propSCL,jd#OnClock, 1)
  outa[comm#propSDA]~
  dira[comm#propSDA]~~
  outa[comm#propSCL]~~
  dira[comm#propSDA]~
  waitcnt(clkfreq/200 + cnt)
           


PRI nm_core_Uart_txInit(bank)
  '/**
  ' * Initialize a Uart transmitter as background task. Its parameters are in vpRambank.
  ' * Called by Uart.start()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)
  

PRI nm_core_Uart_rxInit(bank)
  '/**
  ' * Initialize a Uart receiver as background task. Its parameters are in vpRambank.
  ' * Called by Uart.start()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)


PRI nm_core_Uart_rxRead(bank)
  '/**
  ' * Inform Uart receiver a byte has been grabbed from the buffer, so the reciever can do the handshaking.
  ' * Called by Uart.readByte()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)


PRI nm_core_Callback_returnTwo
  'This appears not to be used

  
PRI nm_core_Timer_init(bank)
  '/**
  ' * Initialize the 32bit counter. The 32bit nmTimer register should increment every 8.68 usec.
  ' * Called by Timer.start()
  ' *
  ' * @param bank The allocated area for the Timer registers.
  ' */


PRI nm_core_Timer_latch(bank)
  '/**
  ' * Latch the 32bit counter in the latch registers.
  ' * Called by Timer.tickLo()
  ' *
  ' * @param bank The allocated area for the Timer registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)
  'This probably needs a lock
  jd.writeRegister(bank + jd#nmLatchTimer1,jd.readRegister(bank + jd#nmTimer1))
  jd.writeRegister(bank + jd#nmLatchTimer2,jd.readRegister(bank + jd#nmTimer2))
  jd.writeRegister(bank + jd#nmLatchTimer3,jd.readRegister(bank + jd#nmTimer3))
  jd.writeRegister(bank + jd#nmLatchTimer4,jd.readRegister(bank + jd#nmTimer4))
  

PRI nm_core_PWM_install(bank) | pin,low,high,duty,scale
  '/**
  ' * Initialize a PWM as background task. Its parameters are in vpRambank.
  ' * Called by PWM.start()
  ' *
  ' * @param bank The allocated area for the PWM registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)
  {{ adapted from object BS2_functions Pub PWM_Set(Pin, Duty, Resolution) | Scale }}
  {{ Produces constant PWM on the specified pin }}
  {{ Needs to calculate scale }}
  if jd.readRegister(bank + jd#nmPWMPort) == 0
    pin := >|jd.readRegister(bank + jd#nmPWMPin)
  else
    pin := 8 + >|jd.readRegister(bank + jd#nmPWMPin)
  low := jd.readRegister(bank + jd#nmPWMlowTime)
  high := jd.readRegister(bank + jd#nmPWMhighTime)
  duty := (high*100) / (low + high)
  if duty == 0                                         ' freq = 0 turns off square wave
     ctra := 0                                          ' Set CTRA/B to 0
     dira[pin]~                                         ' Make pin input
  else
      'scale := 2_147_483_647 / (1<< (Resolution-1))  ' Calculate scale
      ctra[30..26] := %00110      ' Set ctra to DUTY mode
      ctra[5..0] := pin           ' Set ctra's APIN
      frqa := duty * scale        ' Set frqa register
      dira[pin]~~                 ' set direction


PRI nm_core_PWM_updateInternal(bank,highTime,lowTime,highZero,lowZero)
  '/**
  ' * Update the PWM duty cycle. Its parameters are in vpRambank. 
  ' * Called by PWM.update()
  ' *
  ' * @param bank The allocated area for the PWM registers.
  ' * @param highTime the number of 8.68us the waveform should be high.
  ' * @param lowTime the number of 8.68us the waveform should be low.
  ' * @param highZero Boolean value (always true) that tells wether highTime != 0.
  ' * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)
  jd.writeRegister(bank + jd#nmPWMlowTime,lowTime)
  jd.writeRegister(bank + jd#nmPWMhighTime,highTime)
  jd.writeRegister(bank + jd#nmPWMlowZero,lowZero)
  jd.writeRegister(bank + jd#nmPWMhighZero,highZero)
  {{ Need to update ctra and frqa }}
  

PRI nm_core_DAC_install(bank)
  '/**
  ' * Initialize a DAC as background task. Its parameters are in vpRambank.
  ' * Called by DAC.start()
  ' *
  ' * @param bank The allocated area for the DAC registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)
  {{ Need some 8bit DAC routine }}


PRI nm_core_ADC_install(bank)
  '/**
  ' * Initialize an ADC as background task. Its parameters are in vpRambank.
  ' * Called by ADC.start()
  ' *
  ' * @param bank The allocated area for the ADC registers.
  ' */
  'Access the proper register using jd.writeRegister(bank+register)
  {{ Need some 8bit ADC routine }}

  
PRI nm_page_
  'Since the Prop has no pages, simply ignore ???
  
